/*==============================================================================


   Purpose:    example program how to use PCL


   Author:     Rudolf Berrendorf
               University of Applied Sciences Rhein-Sieg
               Computer Science Department
               53754 Sankt Augustin
               rudolf.berrendorf@fh-rhein-sieg.de


   Date:       23.08.2000


   DISCLAIMER

                  PCL - The Performance Counter Library


   Copyright (C) 1999,2000
   Forschungszentrum Juelich GmbH, Federal Republic of Germany.
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are met:

   Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

     - Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.

     - Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.

     - Any publications that result from the use of this software shall
       reasonably refer to the Research Centre's development.

     - All advertising materials mentioning features or use of this software
       must display the following acknowledgement:

           This product includes software developed by Forschungszentrum
           Juelich GmbH, Federal Republic of Germany.

     - Forschungszentrum Juelich GmbH is not obligated to provide the user with
       any support, consulting, training or assistance of any kind with regard
       to the use, operation and performance of this software or to provide
       the user with any updates, revisions or new versions.


   THIS SOFTWARE IS PROVIDED BY FORSCHUNGSZENTRUM JUELICH GMBH "AS IS" AND ANY
   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
   DISCLAIMED. IN NO EVENT SHALL FORSCHUNGSZENTRUM JUELICH GMBH BE LIABLE FOR
   ANY SPECIAL, DIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
   RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
   CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
   CONNECTION WITH THE ACCESS, USE OR PERFORMANCE OF THIS SOFTWARE.

==============================================================================*/


#include <pcl.h>

#include <stdio.h>


/*                                                                            */
/*============================================================================*/
/*                                                                            */


int
main(int argc, char **argv)
{
  PCL_CNT_TYPE i_result[2] = {0, 0};
  PCL_FP_CNT_TYPE fp_result[2];
  int counter_list[2], res;
  unsigned int flags;
  void *descr;


  /* define events to be measured */
  counter_list[0] = PCL_FP_INSTR;
  counter_list[1] = PCL_CYCLES;

  /* define in what mode to count (only user mode) */
  flags = PCL_MODE_USER;

  /* initialize */
  if(PCLinit(&descr) != PCL_SUCCESS)
    {
      printf("unable to initialize\n");
      exit(1);
    }

  /* query for functionality */
  if ((res = PCLquery(descr, counter_list, 2, flags)) != PCL_SUCCESS)
    {
      printf("two events not possible\n");
      exit(1);
    }

  /* start counting */
  if ((res = PCLstart(descr, counter_list, 2, flags)) != PCL_SUCCESS)
    {
      printf("problem with starting two events\n");
      exit(1);
    }

  /* do work */
  /*??? you should include here your piece of code to measure */

  /* stop counting */
  if ((res = PCLstop(descr, i_result, fp_result, 2)) != PCL_SUCCESS)
    printf("problem with stopping two events\n");
  else
    printf("%15.0f FP-instructions in %15.0f cycles\n",
	   (double) i_result[0], (double) i_result[1]);

  /* exit PCL package */
  if(PCLexit(descr) != PCL_SUCCESS)
    {
      printf("problem with exiting PCL\n");
      exit(1);
    }


  return 0;
}


/*                                                                            */
/*============================================================================*/
