/*==============================================================================


   Purpose:    determine MHz rate the processor is running


   Author:     Rudolf Berrendorf
               University of Applied Sciences Rhein-Sieg
               Computer Science Department
               53754 Sankt Augustin
               rudolf.berrendorf@fh-rhein-sieg.de


   Date:       23.02.2000


   DISCLAIMER

                  PCL - The Performance Counter Library


   Copyright (C) 1999, 2000
   Forschungszentrum Juelich GmbH, Federal Republic of Germany.
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are met:

   Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

     - Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.

     - Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.

     - Any publications that result from the use of this software shall
       reasonably refer to the Research Centre's development.

     - All advertising materials mentioning features or use of this software
       must display the following acknowledgement:

           This product includes software developed by Forschungszentrum
           Juelich GmbH, Federal Republic of Germany.

     - Forschungszentrum Juelich GmbH is not obligated to provide the user with
       any support, consulting, training or assistance of any kind with regard
       to the use, operation and performance of this software or to provide
       the user with any updates, revisions or new versions.


   THIS SOFTWARE IS PROVIDED BY FORSCHUNGSZENTRUM JUELICH GMBH "AS IS" AND ANY
   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
   DISCLAIMED. IN NO EVENT SHALL FORSCHUNGSZENTRUM JUELICH GMBH BE LIABLE FOR
   ANY SPECIAL, DIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
   RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
   CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
   CONNECTION WITH THE ACCESS, USE OR PERFORMANCE OF THIS SOFTWARE.

==============================================================================*/


#include "../../determine_mhz.h"

#include <stdio.h>
#include <string.h>
#include <sys/time.h>


/*                                                                            */
/*============================================================================*/
/* calculate the difference of two timeval numbers.                           */


static double
timediff(struct timeval tv2, struct timeval tv1)
{
  return (double)(tv2.tv_sec - tv1.tv_sec)
    + ((double)(tv2.tv_usec - tv1.tv_usec) / 1000000.0);
}


/*                                                                            */
/*============================================================================*/
/* returns MHz-rate of the processor (or -1 if not available)                 */


int
PCL_determine_mhz_rate(void)

{
  int res, mhz;
  char line[256], *s, search_str[] = "cpu MHz";
  FILE *f;


  /* open proc/cpuinfo */
  if((f = fopen("/proc/cpuinfo", "r")) == NULL)
    return -1;

  /* ignore all lines until we reach MHz information */
  res = -1;
  while(fgets(line, 256, f) != NULL)
    {
      if(strstr(line, search_str) != NULL)
	{
	  /* ignore all characters in line up to : */
	  for(s = line; *s && (*s != ':'); ++s)
	    ;
	  /* get MHz number */
	  if(*s && (sscanf(s+1, "%d", &mhz) == 1))
	    {
	      res = mhz;
	      break;
	    }
	}
    }
  

  if(res == -1)
    {
      /* check failed (e.g. old linux version), so try our own test */
      long long clock_value_start, clock_value_end;
      struct timeval tv1, tv2;
      double t_delta;
      
/* time to measure (in sec) */
#define NSECOND 0.01
/* fast assembler code to access the cycle counter (TSC) */
#define rdtsc(valptr) asm volatile ("rdtsc" : \
"=a"(*(unsigned int *)valptr), "=d"(*(((unsigned int *)valptr)+1)))

      /* get actual clock value (to measure cycles) */
      rdtsc(&clock_value_start);

      /* do work */
      gettimeofday(&tv1, NULL);
      do
	{
	  gettimeofday(&tv2, NULL);
	}
      while((t_delta = timediff(tv2, tv1)) < NSECOND);
      
      /* get actual clock value */
      rdtsc(&clock_value_end);
      
      res = (double)(clock_value_end - clock_value_start)/t_delta/1000000.0;
    }

  return res;
}


/*============================================================================*
 *                             that's all folks                               *
 *============================================================================*/
