/*==============================================================================


   Purpose:    low level routines for PowerPC


   Author:     Rudolf Berrendorf
               University of Applied Sciences Rhein-Sieg
               Computer Science Department
               53754 Sankt Augustin
               rudolf.berrendorf@fh-rhein-sieg.de


   last mod:   01.09.2000


   DISCLAIMER

                  PCL - The Performance Counter Library


   Copyright (C) 2000
   Forschungszentrum Juelich GmbH, Federal Republic of Germany.
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are met:

   Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

     - Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.

     - Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.

     - Any publications that result from the use of this software shall
       reasonably refer to the Research Centre's development.

     - All advertising materials mentioning features or use of this software
       must display the following acknowledgement:

           This product includes software developed by Forschungszentrum
           Juelich GmbH, Federal Republic of Germany.

     - Forschungszentrum Juelich GmbH is not obligated to provide the user with
       any support, consulting, training or assistance of any kind with regard
       to the use, operation and performance of this software or to provide
       the user with any updates, revisions or new versions.


   THIS SOFTWARE IS PROVIDED BY FORSCHUNGSZENTRUM JUELICH GMBH "AS IS" AND ANY
   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
   DISCLAIMED. IN NO EVENT SHALL FORSCHUNGSZENTRUM JUELICH GMBH BE LIABLE FOR
   ANY SPECIAL, DIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
   RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
   CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
   CONNECTION WITH THE ACCESS, USE OR PERFORMANCE OF THIS SOFTWARE.
  
==============================================================================*/

#include "../../defines.h"
#include "../../pcl.h"
#include "hw_driver_defs.h"
#include "../../pcl_hw_driver.h"
#include "../../hw_driver_global_defs.h"
#include "../../state_descr.h"

#include <stdlib.h>
#include <pmapi.h>
#include <stdio.h>


/*                                                                            */
/*----------------------------------------------------------------------------*/
/* local variables                                                            */

/* commands are actually indices into the event name table *pmapi_names[]
   below
*/

/* invalid entry-data if event is not available */
#define PCL_INVALID    0, {{{-1,-1,-1,-1,-1,-1,-1,-1,-1},{0,0,0,0,0,0,0,0,0}}}

/* PowerPC 604 (2 counter + elapsed counter)
   and PowerPC604 (4 counters + elapsed counter)
*/
/* Entries in this table are indices to the string table "pmapi_names" below */
static const PCL_event_descr_t event_table_604[PCL_MAX_EVENT] =
{
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1CACHE_HIT */
  {PCL_AVAIL,COMPOSED,{{{4,18,-1,-1,-1,-1,-1,-1,-1},{OP_ADD,OP_ADD,-1,-1,-1,-1,-1,-1,-1}}}}, /* PCL_L1CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1DCACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1DCACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{-1,18,-1,-1,-1,-1,-1,-1,-1},{-1,OP_ADD,-1,-1,-1,-1,-1,-1,-1}}}}, /* PCL_L1DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1ICACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{4,-1,-1,-1,-1,-1,-1,-1,-1},{OP_ADD,-1,-1,-1,-1,-1,-1,-1,-1}}}}, /* PCL_L1ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2CACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2DCACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2DCACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2ICACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_TLB_HIT */
  {PCL_AVAIL,COMPOSED,{{{19,20,-1,-1,-1,-1,-1,-1,-1},{OP_ADD,OP_ADD,-1,-1,-1,-1,-1,-1,-1}}}}, /* PCL_TLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_ITLB_HIT */
  {PCL_AVAIL,NOT_COMP,{{{-1,20,-1,-1,-1,-1,-1,-1,-1},{-1,OP_ADD,-1,-1,-1,-1,-1,-1,-1}}}}, /* PCL_ITLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_DTLB_HIT */
  {PCL_AVAIL,NOT_COMP,{{{19,-1,-1,-1,-1,-1,-1,-1,-1},{OP_ADD,-1,-1,-1,-1,-1,-1,-1,-1}}}}, /* PCL_DTLB_MISS */
  {PCL_AVAIL,NOT_COMP,{{{6,6,-1,-1,-1,-1,-1,-1,-1},{OP_ADD,OP_ADD,-1,-1,-1,-1,-1,-1,-1}}}}, /* PCL_CYCLES */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_ELAPSED_CYCLES */
  {PCL_AVAIL,NOT_COMP,{{{21,-1,-1,-1,-1,-1,-1,-1,-1},{OP_ADD,-1,-1,-1,-1,-1,-1,-1,-1}}}}, /* PCL_INTEGER_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{22,-1,-1,-1,-1,-1,-1,-1,-1},{OP_ADD,-1,-1,-1,-1,-1,-1,-1,-1}}}}, /* PCL_FP_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{-1,12,-1,-1,-1,-1,-1,-1,-1},{-1,OP_ADD,-1,-1,-1,-1,-1,-1,-1}}}}, /* PCL_LOAD_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_STORE_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_LOADSTORE_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{14,14,-1,-1,-1,-1,-1,-1,-1},{OP_ADD,OP_ADD,-1,-1,-1,-1,-1,-1,-1}}}}, /* PCL_INSTR */
  {PCL_AVAIL,COMPOSED,{{{23,16,-1,-1,-1,-1,-1,-1,-1},{OP_SUB,OP_ADD,-1,-1,-1,-1,-1,-1,-1}}}}, /* PCL_JUMP_SUCCESS */
  {PCL_AVAIL,NOT_COMP,{{{23,-1,-1,-1,-1,-1,-1,-1,-1},{OP_ADD,-1,-1,-1,-1,-1,-1,-1,-1}}}}, /* PCL_JUMP_UNSUCCESS */
  {PCL_AVAIL,NOT_COMP,{{{-1,16,-1,-1,-1,-1,-1,-1,-1},{-1,OP_ADD,-1,-1,-1,-1,-1,-1,-1}}}}, /* PCL_JUMP */
  {PCL_AVAIL,NOT_COMP,{{{-1,24,-1,-1,-1,-1,-1,-1,-1},{-1,OP_ADD,-1,-1,-1,-1,-1,-1,-1}}}}, /* PCL_ATOMIC_SUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_ATOMIC_UNSUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_ATOMIC */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_STALL_INTEGER */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_STALL_FP */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_STALL_JUMP */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_STALL_LOAD */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_STALL_STORE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_STALL */
  {PCL_AVAIL,COMPOSED,{{{22,6,-1,-1,-1,-1,-1,-1,-1},{OP_FP_1|OP_FP_2|OP_FP_STORE,OP_FP_2|OP_FP_DIV_MHZ,-1,-1,-1,-1,-1,-1,-1}}}}, /* PCL_MFLOPS */
  {PCL_AVAIL,COMPOSED,{{{14,6,-1,-1,-1,-1,-1,-1,-1},{OP_FP_1|OP_FP_2|OP_FP_STORE,OP_FP_2|OP_FP_DIV,-1,-1,-1,-1,-1,-1,-1}}}}, /* PCL_IPC */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1DCACHE_MISSRATE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2DCACHE_MISSRATE */
  {PCL_UNAVAIL, PCL_INVALID}                 /* PCL_MEM_FP_RATIO */
};


/* Power3 aka 630 and Power3-II aka 630p */
/* Entries in this table are indices to the string table "pmapi_names" below */
static const PCL_event_descr_t event_table_630[PCL_MAX_EVENT] =
{
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1CACHE_HIT */
  {PCL_AVAIL,COMPOSED,{{{0,1,-1,-1,-1,-1,-1,-1,-1},{OP_ADD,OP_ADD,-1,-1,-1,-1,-1,-1,-1}}}}, /* PCL_L1CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1DCACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1DCACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{2,2,-1,-1,-1,2,-1,-1,-1},{OP_ADD,OP_ADD,-1,-1,-1,OP_ADD,-1,-1,-1}}}}, /* PCL_L1DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1ICACHE_READWRITE */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,-1,-1,3,-1,-1,-1,-1},{-1,-1,-1,-1,OP_ADD,-1,-1,-1,-1}}}}, /* PCL_L1ICACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{4,-1,-1,-1,-1,-1,4,-1,-1},{OP_ADD,-1,-1,-1,-1,-1,OP_ADD,-1,-1}}}}, /* PCL_L1ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2CACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2DCACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2DCACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2ICACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_TLB_HIT */
  {PCL_AVAIL,NOT_COMP,{{{5,-1,-1,-1,-1,-1,-1,5,-1},{OP_ADD,-1,-1,-1,-1,-1,-1,OP_ADD,-1}}}}, /* PCL_TLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_ITLB_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_ITLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_DTLB_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_DTLB_MISS */
  {PCL_AVAIL,NOT_COMP,{{{6,6,6,6,6,6,6,6,-1},{OP_ADD,OP_ADD,OP_ADD,OP_ADD,OP_ADD,OP_ADD,OP_ADD,OP_ADD,-1}}}}, /* PCL_CYCLES */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_ELAPSED_CYCLES */
  {PCL_AVAIL,COMPOSED,{{{-1,-1,-1,-1,-1,7,8,9,-1},{-1,-1,-1,-1,-1,OP_ADD,OP_ADD,OP_ADD,-1}}}}, /* PCL_INTEGER_INSTR */
  {PCL_AVAIL,COMPOSED,{{{-1,10,-1,-1,11,-1,-1,-1,-1},{-1,OP_ADD,-1,-1,OP_ADD,-1,-1,-1,-1}}}}, /* PCL_FP_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{12,-1,-1,12,-1,-1,-1,-1,-1},{OP_ADD,-1,-1,OP_ADD,-1,-1,-1,-1,-1}}}}, /* PCL_LOAD_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,13,-1,-1,-1,-1,-1,-1},{-1,-1,OP_ADD,-1,-1,-1,-1,-1,-1}}}}, /* PCL_STORE_INSTR */
  {PCL_AVAIL,COMPOSED,{{{-1,-1,13,12,-1,-1,-1,-1,-1},{-1,-1,OP_ADD,OP_ADD,-1,-1,-1,-1,-1}}}}, /* PCL_LOADSTORE_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{14,14,14,14,-1,-1,-1,-1,-1},{OP_ADD,OP_ADD,OP_ADD,OP_ADD,-1,-1,-1,-1,-1}}}}, /* PCL_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{-1,15,-1,-1,-1,-1,-1,-1,-1},{-1,OP_ADD,-1,-1,-1,-1,-1,-1,-1}}}}, /* PCL_JUMP_SUCCESS */
  {PCL_AVAIL,COMPOSED,{{{-1,15,-1,16,-1,-1,-1,-1,-1},{-1,OP_SUB,-1,OP_ADD,-1,-1,-1,-1,-1}}}}, /* PCL_JUMP_UNSUCCESS */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,-1,16,-1,-1,-1,-1,-1},{-1,-1,-1,OP_ADD,-1,-1,-1,-1,-1}}}}, /* PCL_JUMP */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_ATOMIC_SUCCESS */
  {PCL_AVAIL,NOT_COMP,{{{-1,17,-1,-1,-1,-1,-1,-1,-1},{-1,OP_ADD,-1,-1,-1,-1,-1,-1,-1}}}}, /* PCL_ATOMIC_UNSUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_ATOMIC */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_STALL_INTEGER */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_STALL_FP */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_STALL_JUMP */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_STALL_LOAD */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_STALL_STORE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_STALL */
  {PCL_AVAIL,COMPOSED,{{{-1,10,-1,-1,11,-1,-1,6,-1},{-1,OP_FP_1|OP_FP_2|OP_FP_STORE,-1,-1,OP_FP_2|OP_FP_ADD,-1,-1,OP_FP_2|OP_FP_DIV_MHZ,-1}}}}, /* PCL_MFLOPS */
  {PCL_AVAIL,COMPOSED,{{{14,6,-1,-1,-1,-1,-1,-1,-1},{OP_FP_1|OP_FP_2|OP_FP_STORE,OP_FP_2|OP_FP_DIV,-1,-1,-1,-1,-1,-1,-1}}}}, /* PCL_IPC */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1DCACHE_MISSRATE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2DCACHE_MISSRATE */
  {PCL_UNAVAIL, PCL_INVALID}                 /* PCL_MEM_FP_RATIO */
};


/* This is a list of event names for pmapi.
   Commands are actually indices into this table.
 */
static const char *pmapi_names[] =
{
  "PM_LD_MISS_L2HIT",  /* 0 */
  "PM_ST_MISS", /* 1 */
  "PM_LD_MISS_L2HIT", /* 2 */
  "PM_IC_HIT", /* 3 */
  "PM_IC_MISS", /* 4 */
  "PM_TLB_MISS", /* 5 */
  "PM_CYC", /* 6 */
  "PM_FXU2_PROD_RESULT", /* 7 */
  "PM_FXU0_PROD_RESULT", /* 8 */
  "PM_FXU1_PROD_RESULT", /* 9 */
  "PM_FPU1_CMPL", /* 10 */
  "PM_FPU0_CMPL", /* 11 */
  "PM_LD_CMPL", /* 12 */
  "PM_ST_CMPL", /* 13 */
  "PM_INST_CMPL", /* 14 */
  "PM_BR_PRED", /* 15 */
  "PM_BR_CMPL", /* 16 */
  "PM_ST_COND_FAIL", /* 17 */
  "PM_DC_MISS", /* 18 */
  "PM_DTLB_MISS", /* 19 */
  "PM_ITLB_MISS", /* 20 */
  "PM_FXU_CMPL", /* 21 */
  "PM_FPU_CMPL", /* 22 */
  "PM_BR_MPRED", /* 23 */
  "PM_RESRV_CMPL" /* 24 */
};


/*                                                                            */
/*============================================================================*/
/* returns relative index of event code for counter ctr                       */
/* this is because PMAPI has a totally different way to specify event codes.  */
/* We have to translate our scheme into theirs.                               */


static int
search_event(
	     PCL_state_descr *descr,
	     pm_info_t pminfo,
	     int ctr,
	     PCL_DRIVER_COMMAND_TYPE command
	     )

{
  int i, j;
  pm_events_t *wevp;


  /* search in pminfo with the name for the correct index */
  wevp = pminfo.list_events[ctr];
  for(j = 0; j < pminfo.maxevents[ctr]; ++j, ++wevp)
    if(!strcmp(wevp->short_name, pmapi_names[command]))
      return wevp->event_id;

  /* not found */
  return -1;
}


/*                                                                            */
/*============================================================================*/
/* display interesting parts of a pminfo data structure                       */


static void
display_pminfo(
	       pm_info_t pminfo
	       )

{
  int i,j;
  pm_events_t	*wevp;

	
  printf("pminfo:\n\tcounters:\t%d\n", pminfo.maxpmcs);
  for(i = 0; i < pminfo.maxpmcs; ++i)
    {
      printf("\tcounter %d: %d\n", i, pminfo.maxevents[i]);
      wevp = pminfo.list_events[i];
      for(j = 0; j < pminfo.maxevents[i]; ++j, ++wevp)
	printf("%s ", wevp->short_name);
      printf("\n");
    }
}


/*                                                                            */
/*============================================================================*/
/* initialization                                                             */

int
PCL_driver_open(
		PCL_DESCR_TYPE d /* I: descriptor */
		)

{
  int rc;
  PCL_state_descr *descr;
  HW_state_descr *hw_descr;


  /* get state descriptor */
  descr = d;
  hw_descr = &descr->hw_state_descr;

  /* set allowed count mode */
  descr->allowed_count_mode = PCL_MODE_USER_SYSTEM;

  /* initialize PMAPI */
  if((rc = pm_init(PM_VERIFIED|PM_UNVERIFIED|PM_CAVEAT, &hw_descr->pminfo)) > 0)
    return PCL_FAILURE;

  /* check processor type */
  if((!strcmp(hw_descr->pminfo.proc_name, "PowerPC 604"))
     || (!strcmp(hw_descr->pminfo.proc_name, "PowerPC 604e")))
    {
      descr->event_table = event_table_604;
      hw_descr->cycle_counter_index = 2;
    }
  else if((!strcmp(hw_descr->pminfo.proc_name, "POWER3"))
	  || (!strcmp(hw_descr->pminfo.proc_name, "POWER3-II")))
    {
      descr->event_table = event_table_630;
      hw_descr->cycle_counter_index = 8;
    }
  else
    {
      fprintf(stderr, "PCL: unknown processor (%s)\n",
	      hw_descr->pminfo.proc_name);
      return PCL_FAILURE;
    }

#if 0
  /* display pminfo structure */
  display_pminfo(hw_descr->pminfo);
#endif

  /* everything ok */
  return PCL_SUCCESS;
}


/*                                                                            */
/*============================================================================*/
/* Starts hardware counters.                                                  */


int
PCL_driver_start(
		 PCL_DESCR_TYPE d, /* I: descriptor */
		 int max_counter,
		 PCL_BIT_MASK_TYPE counter_used_mask,
		 PCL_DRIVER_COMMAND_TYPE *commands,
		 unsigned int count_mode
		 )

{
  int i, rc;
  PCL_BIT_MASK_TYPE used_mask;
  pm_prog_t prog;
  PCL_state_descr *descr;
  HW_state_descr *hw_descr;


  /* get state descriptor */
  descr = d;
  hw_descr = &descr->hw_state_descr;

  /* set up used mask */
  used_mask = counter_used_mask & PCL_mask[max_counter-1];

  /* start with clean mode */
  prog.mode.w = 0;

  /* count in user mode? */
  if(count_mode & PCL_MODE_USER)
    prog.mode.b.user = 1;

  /* count in kernel mode? */
  if(count_mode & PCL_MODE_SYSTEM)
    prog.mode.b.kernel = 1;

  /* start counting immediately */
  prog.mode.b.count = 1;

  /* init counters to not count anything */
  for (i = 0; i < hw_descr->pminfo.maxpmcs; i++)
    prog.events[i] = COUNT_NOTHING;
 
  /* Extract selected events */
  for(i = 0; i < max_counter; ++i)
    if((i != hw_descr->cycle_counter_index) && (IS_USED(used_mask, i)))
      {
	prog.events[i] = search_event(descr, hw_descr->pminfo, i, commands[i]);
	if(prog.events[i] == -1)
	  return PCL_FAILURE;
      }

  /* create and init a pm context for this thread */
  if ((rc = pm_set_program_mythread(&prog)) > 0)
    return PCL_FAILURE;

#if 0
  if(IS_USED(used_mask, hw_descr->cycle_counter_index))
    /* get actual clock value (to measure cycles) */
    hw_descr->clock_value = ???();
#endif

  /* everything ok */
  return PCL_SUCCESS;
}


/*                                                                            */
/*============================================================================*/
/* Read hardware counters                                                     */

int
PCL_driver_read(
		PCL_DESCR_TYPE d, /* I: descriptor */
		int max_counter,
		PCL_BIT_MASK_TYPE counter_used_mask,
		PCL_CNT_TYPE *counter_values
		)

{
  int i, rc;
  PCL_BIT_MASK_TYPE used_mask;
  pm_data_t perfdata;
  PCL_state_descr *descr;
  HW_state_descr *hw_descr;


  /* get state descriptor */
  descr = d;
  hw_descr = &descr->hw_state_descr;

  /* set up used mask */
  used_mask = counter_used_mask & PCL_mask[max_counter-1];

#if 0
  /* get actual cycle count */
  if(IS_USED(used_mask, hw_descr->cycle_counter_index))
    counter_values[hw_descr->cycle_counter_index] = ???() - 
      hw_descr->clock_value;
#endif

  /* read pm data for this thread */
  if ((rc = pm_get_data_mythread(&perfdata)) > 0)
    return PCL_FAILURE;

  for(i = 0; i < max_counter; ++i)
    if(IS_USED(counter_used_mask, i))
      counter_values[i] = perfdata.accu[i];

  /* everything ok */
  return PCL_SUCCESS;
}


/*                                                                            */
/*============================================================================*/
/*                                                                            */


int
PCL_driver_stop(
		PCL_DESCR_TYPE d /* I: descriptor */
		)

{
  /* stop counting for this thread */
  if (pm_stop_mythread() > 0) 
    return PCL_FAILURE;

  /* delete pm context */
  if(pm_delete_program_mythread() > 0)
    return PCL_FAILURE;

  /* everything ok */
  return PCL_SUCCESS;
}


/*                                                                            */
/*============================================================================*/
/* close                                                                      */

int
PCL_driver_close(
		 PCL_DESCR_TYPE d /* I: descriptor */
		 )

{
  /* everything ok */
  return PCL_SUCCESS;
}


/*============================================================================*
 *                             that's all folks                               *
 *============================================================================*/
